@@ -4,7 +4,9 @@ from configdict import ConfigDict
 modulename = 'csoundengine.engine'
 
 logger = logging.getLogger('csoundengine')
-
+_handler = logging.StreamHandler()
+_handler.setFormatter(logging.Formatter("*** csoundengine: %(message)s\n"))
+logger.addHandler(_handler)
 
 # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
 #                  CONFIG                   #
@@ -15,13 +17,13 @@ _ = config.addKey
 
 _('sr', 0,
   choices=(0, 22050, 44100, 48000, 88200, 96000),
-  doc='0 indicates the default sr for the backend')
+  doc='samplerate - 0=default sr for the backend')
 _('rec.sr', 44100,
   choices=(44100, 48000, 88200, 96000, 192000),
-  doc='default sample rate when rendering')
+  doc='default samplerate when rendering')
 _('nchnls', 0,
   range=(0, 128),
-  doc='Number of output channels to use. 0 = default for used device')
+  doc='Number of output channels. 0=default for used device')
 _('nchnls_i', 0,
   range=(0, 128),
   doc='Number of input channels to use. 0 = default for used device')
@@ -30,15 +32,14 @@ _('ksmps', 64,
   doc="corresponds to csound's ksmps")
 _('rec.ksmps', 64,
   choices=(16, 32, 64, 128, 256),
-  doc="corresponds to csound's ksmps, used when rendering")
+  doc="samples per cycle when rendering")
 _('rec.sample_format', 'float',
   choices=(16, 24, 32, 'float'),
   doc="Sample format used when rendering")
 _('buffersize', 0,
   doc="-b value. 0=determine buffersize depending on ksmps & backend")
 _('numbuffers', 0,
-  doc="determines the -B value as a multiple of the buffersize (-b). "
-      "0 = determine numbuffers depending on the backend")
+  doc="determines the -B value as a multiple of the buffersize. 0=auto")
 _('linux.backend', 'jack',
   choices=('jack', 'pa_cb', 'pa_bl', 'pulse', 'alsa')),
 _('macos.backend', 'pa_cb',
@@ -46,35 +47,33 @@ _('macos.backend', 'pa_cb',
 _('windows.backend', 'pa_cb',
   choices=('pa_cb', 'pa_bl'))
 _('fallback_backend', 'pa_cb',
-  choices=('pa_cb', 'pa_bl'))
+  choices=('pa_cb', 'pa_bl', ''),
+  doc="Fallback backend if the preferred backend is not available.")
 _('A4', 442,
-  range=(410, 460))
+  range=(410, 460),
+  doc="Frequency for A4")
 _('check_pargs', False,
   doc='Check number of pargs passed to instr')
 _('fail_if_unmatched_pargs', False,
   doc='Fail if the # of passed pargs doesnt match the # of pargs'),
-_('wait_poll_interval', 0.020,
-  doc='seconds to wait when polling for a synth to finish')
 _('set_sigint_handler', True,
   doc='Set a sigint handler to prevent csound crash with CTRL-C')
 _('generalmidi_soundfont', '')
 _('suppress_output', True,
   doc='Supress csound´s debugging information')
 _('unknown_parameter_fail_silently', True,
-  doc='Do not raise an Error if a synth is asked to set an unknown parameter')
+  doc='Do not raise if a synth tries to set an unknown parameter')
 _('define_builtin_instrs', True,
-  doc="Whenever a Session is created, it will have all builtin "
-      "instruments defined")
+  doc="If True, a Session with have all builtin instruments defined")
 _('sample_fade_time', 0.05,
-  doc="Default fade time when playing samples via a Session")
-_("prefer_udp", False,
-  doc="If true and a UDP server has been defined, prefer UDP for communication"
-      "instead of the API")
+  doc="Fade time when playing samples via a Session")
+_("prefer_udp", True,
+  doc="If true and a server was defined prefer UDP over the API for communication")
 _("start_udp_server", False,
-  doc="If True, start an engine with udp communication support")
-_('autostart_engine', True,
-  doc="If True, autostart an engine as needed whenever creating a Session")
-_('associated_table_min_size', 16)
-_('num_audio_buses', 64)
+  doc="Start an engine with udp communication support")
+_('associated_table_min_size', 16,
+  doc="Min. size of the param table associated with a synth")
+_('num_audio_buses', 64,
+  doc="Num. of audio buses in an Engine/Session")
 
 config.load()
