@@ -1,85 +1,39 @@
 from __future__ import annotations
 from string import Template
+from functools import lru_cache
+from typing import List, Dict, Any
+import re
 
+# In all templates we make the difference between substitutions which
+# are constant (with the form ${subst} and substitutions which respond
+# to the configuration of a specific engine / offline renderer ({subst})
 
 _orc = r"""
-sr     = {sr}
-ksmps  = {ksmps}
-nchnls = {nchnls}
-nchnls_i = {nchnls_i}
+sr     = ${sr}
+ksmps  = ${ksmps}
+nchnls = ${nchnls}
+nchnls_i = ${nchnls_i}
 0dbfs  = 1
-A4     = {a4}
+A4     = ${a4}
 
-{includes}
+${includes}
 
 gi__responses ftgen  ${responses}, 0, ${numtokens}, -2, 0
 gi__subgains  ftgen  ${subgains},  0, 100, -2, 1
-ga__buses[] init {numAudioBuses}
-gi__busrefs ftgen ${busrefs}, 0, {numAudioBuses}, -2, 0
-
-{globalcode}
+gi__tokenToInstrnum ftgen ${tokenToInstrnum}, 0, ${maxNumInstrs}, -2, 0
 
+${globalcode}
 
 ; ---------------------------------
 ;          builtin-instruments
 ; ---------------------------------
 
-instr _init
+instr __init
     ftset gi__subgains, 1
-    prints "<<< Init done >>>\n"
+    ;; prints "<<< Init done >>>\n"
     turnoff
 endin
 
-instr ${highestInstrnum}
-    ; this is used to prevent a crash when an opcode is defined as part 
-    ; of globalcode, and later on an instr is defined with a high instrnum
-    turnoff
-endin
-
-instr ${cleanbuses}
-    zeroarray ga__buses
-endin
-
-opcode _bususe, 0, i
-    iidx xin
-    ; prints "using bus %d \n", iidx
-    ival tab_i iidx, gi__busrefs
-    tabw_i ival+1, iidx, gi__busrefs
-    atstop "_busdec", 0, -1, iidx
-endop
-
-instr _busdec
-    iidx = p4
-    ival tab_i iidx, gi__busrefs
-    tabw_i ival-1, iidx, gi__busrefs
-endin
-
-opcode busnew, i, 0
-    iidx ftfind gi__busrefs, 0
-    if iidx < 0 then
-        initerror "Out of buses"
-    endif
-    xout iidx    
-endop
-
-opcode busmix, 0, ia
-    iidx, asig xin
-    _bususe(iidx)
-    ga__buses[iidx] = ga__buses[iidx] + asig
-endop
-
-opcode busout, 0, ia
-    iidx, asig xin
-    _bususe(iidx)
-    ga__buses[iidx] = asig
-endop
-
-opcode busin, a, i
-    iidx xin
-    _bususe(iidx)
-    xout ga__buses[iidx]
-endop
-
 opcode _panstereo, aa, aak
     ; kpos: 0-1
     a0, a1, kpos xin
@@ -110,8 +64,9 @@ instr ${nstrnum}
         tabw_i inum, itoken, gi__responses
         outvalue "__sync__", itoken
     endif
+    turnoff
 endin
-
+     
 instr ${tabwrite}
     itab = p4
     iidx = p5
@@ -127,17 +82,6 @@ instr ${chnset}
     turnoff
 endin
 
-; this is used to fill a table with the given pargs
-instr ${filltable}
-    itoken, ifn, itabidx, ilen passign 4
-    iArgs[] passign 8, 8+ilen
-    copya2ftab iArgs, ifn, itabidx
-    if itoken > 0 then
-        outvalue "__sync__", itoken
-    endif
-    turnoff
-endin
-
 instr ${pingback}
     itoken = p4
     outvalue "__sync__", itoken
@@ -160,8 +104,6 @@ instr ${maketable}
     iempty = p7
     isr = p8
     inumchannels = p9
-    prints ">>> instr maketable\n"
-
     if (iempty == 1) then
         ifn ftgen itabnum, 0, ilen, -2, 0
     else
@@ -224,46 +166,6 @@ instr ${automateTableViaTable}
     tablew ky, iargidx, iargtab
 endin
 
-instr ${automate}
-    iargtab = p4
-    iargidx = p5
-    inumpairs = p6
-    imode = p7   ; 0 = linear, 1 = cos, 2 = lag
-    iparam = p8
-    idataidx = 9
-
-    if ftexists:i(iargtab) == 0 then
-        initerror sprintf("Table %d doesn't exist", iargtab)
-    endif
-    if ftlen(iargtab) <= iargidx then
-        initerror sprintf("Table too small (%d <= %d)", ftlen(iargtab), iargidx)
-    endif
-
-    ilen = inumpairs*2
-    iValues[] passign idataidx, idataidx+ilen
-    ; WTF: slicearray's end index is inclusive
-    iXs[] slicearray iValues, 0, ilen-1, 2
-    iYs[] slicearray iValues, 1, ilen-1, 2
-    kt timeinsts
-    if imode == 0 then
-        ky bpf kt, iXs, iYs
-    elseif imode == 1 then
-        ky bpfcos kt, iXs, iYs
-    elseif imode == 2 then
-        ky bpf kt, iXs, iYs
-        ilagtime = max(iparam, 0.05)
-        ky = lag(ky, ilagtime)
-    else
-        throwerror "init", sprintf("imode %d not supported", imode)
-    endif
-    if ftexists:k(iargtab) == 0 then
-        throwerror "warning", \
-            sprintf("automate: dest table (%d) was freed, stopping", iargtab)
-        turnoff
-    endif
-    tablew ky, iargidx, iargtab
-endin
-
 instr ${uniqinstance}
     itoken = p4
     ip1    = p5
@@ -302,16 +204,16 @@ instr ${playsndfile}
     Spath  = p4
     kgain  = p5
     kspeed = p6
-    ichan  = p6
+    ichan  = p7
     ifade  = p8
     idur = filelen(Spath)
     know init 0
-    iwsize = ispeed == 1 ? 1 : 4
+    iwsize = 4   ; cubic interpolation
     aouts[] diskin2 Spath, kspeed, 0, 0, 0, iwsize
     inumouts = lenarray(aouts)
     ichans[] genarray ichan, ichan+inumouts-1
     aenv linsegr 0, ifade, 1, ifade, 0
-    aenv *= intrp(kgain)
+    aenv *= interp(kgain)
     aouts = aouts * aenv
     poly0 inumouts, "outch", ichans, aouts
     know += 1/kspeed
@@ -325,28 +227,34 @@ instr ${readSndfile}
     Spath = strget(p5)
     itab = p6
     ichan = p7
-    itab ftgen 0, 0, -1, Spath, 0, 0, ichan
-    tabw_i itab, itoken, gi__responses
+    itab2 ftgen itab, 0, 0, -1, Spath, 0, 0, ichan
+    ; prints ">>> Spath: %s, itab: %d, itab2: %d \n", Spath, itab, itab2
+    tabw_i itab2, itoken, gi__responses
     outvalue "__sync__", itoken
 endin
 
 instr ${playgen1}
-    ;             4     5      6       7     8     9          10
-    ;             gain, speed, tabnum, chan, fade, starttime, gaingroup
-    pset 0, 0, 0, 0,    1,     1,      1,    0.05, 0,         0
+    ;             4     5      6       7     8     9          10        11
+    ;             gain, speed, tabnum, chan, fade, starttime, gaingroup lagtime
+    pset 0, 0, 0, 1,    1,     1,      1,    0.05, 0,         0,        0.01
     kgain = p4
     kspeed = p5
     ksampsplayed = 0
-    itabnum, ichan, ifade, ioffset, igaingroup passign 6
+    itabnum, ichan, ifade, ioffset, igaingroup, ilagtime passign 6
+    ifade = max(ifade, ksmps/sr*2)
     inumsamples = nsamp(itabnum)
     itabsr = ftsr(itabnum)
+    if itabsr <= 0 then
+        initerror sprintf("Could not determine sr for table %d", itabnum)
+    endif
     istartframe = ioffset * itabsr
     ksampsplayed += ksmps * kspeed
     ; ar[] loscilx xamp, kcps, ifn, iwsize, ibas, istrt
-    aouts[] loscilx kgain, kspeed, itabnum, 4, 1, istartframe
+    aouts[] loscilx 1, kspeed, itabnum, 4, 1, istartframe
     aenv = linsegr:a(0, ifade, 1, ifade, 0)
-    ksubgain = table:k(igaingroup, gi__subgains)
-    aenv *= a(ksubgain)
+    kgain *= table:k(igaingroup, gi__subgains)
+    again = lag:a(a(kgain), ilagtime)
+    aenv *= again
     aouts = aouts * aenv
     inumouts = lenarray(aouts)
     kchan = 0
@@ -396,10 +304,10 @@ instr ${testaudio}
     imode passign 4
     kchan init -1
     if imode == 0 then
-        ; prints "Testaudio: pink noise mode\n"
+        prints "Testaudio: pink noise mode\n"
         a0 pinker
     elseif imode == 1 then
-        ; prints "Testaudio: sine tone mode\n"
+        prints "Testaudio: sine tone mode\n"
         a0 oscili 0.1, 1000
     endif
     kswitch metro 1
@@ -410,32 +318,136 @@ instr ${testaudio}
     endif
 endin
 
-schedule "_init", 0, 1
+instr ${tableInfo}
+    itabnum, itok1, itok2, itok3, itok4 passign 4
+    if ftexists(itabnum) == 0 then
+        initerror sprintf("Table %d does not exist", itabnum)
+    endif
+    isr ftsr itabnum
+    ichnls ftchnls itabnum
+    inumframes nsamp itabnum
+    ilen ftlen itabnum
+    
+    tabw_i isr, itok1, gi__responses
+    tabw_i ichnls, itok2, gi__responses
+    tabw_i inumframes, itok3, gi__responses
+    tabw_i ilen, itok4, gi__responses
+    outvalue "__sync__", itok1
+endin
+
+schedule "__init", 0, 1
 
 """
 
-_instrNames = [
-    'turnoff',
-    'chnset',
-    'filltable',
-    'freetable',
-    'maketable',
-    'playsndfile',
-    'uniqinstance',
-    'scheduniq',
-    'strset',
-    'playgen1',
-    'ftsetparams',
-    'pwrite',
-    'automate',
-    'testaudio',
-    'tabwrite',
-    'automateTableViaTable',
-    'automatePargViaTable',
-    'readSndfile',
-    'pingback',
-    'nstrnum'
-]
+_busOrc = r'''
+
+; The actual buses
+ga__buses[]   init ${numAudioBuses}
+
+; This table keeps track of the number of references a bus has
+gi__busrefs ftgen 0, 0, ${numAudioBuses}, -2, 0
+
+; A pool of bus indexes
+gi__buspool pool_gen ${numAudioBuses}
+
+; A dict mapping bustoken to bus number
+gi__bustoken2num dict_new "int:float"
+
+chn_k "_busTokenCount", 3
+
+instr __businit
+    chnset 0, "_busTokenCount"
+    turnoff
+endin
+
+instr ${clearbuses}
+    zeroarray ga__buses
+endin
+
+opcode busassign, i, j
+    itoken xin
+    if itoken == -1 then
+        itoken chnget "_busTokenCount"
+    endif
+    ibus pool_pop gi__buspool, -1
+    if ibus == -1 then
+        initerror "busassign failed, out of buses"
+    endif
+    dict_set gi__bustoken2num, itoken, ibus
+    chnset itoken+1, "_busTokenCount"
+    xout ibus
+endop
+
+instr _busassign
+    itoken = p4
+    ibus busassign itoken
+    turnoff
+endin
+
+instr _busrelease
+    itoken = p4
+    ibus dict_get gi__bustoken2num, itoken, -1
+    if ibus == -1 then
+        initerror sprintf("itoken %d has no bus assigned to it", itoken)
+    endif
+    irefs tab_i ibus, gi__busrefs
+    if irefs <= 1 then
+        if pool_isfull:i(gi__buspool) == 1 then
+            initerror "Bus pool is full!"
+        endif
+        pool_push gi__buspool, ibus
+        dict_del gi__bustoken2num, itoken
+        tabw_i 0, ibus, gi__busrefs
+    else   
+        tabw_i irefs-1, ibus, gi__busrefs
+    endif
+    turnoff    
+endin
+
+opcode _bususe, i, i
+    itoken xin
+    ibus dict_get gi__bustoken2num, itoken, -1
+    if ibus == -1 then
+        ibus = busassign(itoken)
+    endif
+    irefs tab_i ibus, gi__busrefs
+    tabw_i irefs+1, ibus, gi__busrefs
+    atstop "_busrelease", 0, 0.1, itoken
+    xout ibus
+endop
+
+opcode busin, a, i
+    itoken xin
+    ibus = _bususe(itoken)
+    xout ga__buses[ibus]
+endop
+
+opcode busout, 0, ia
+    itoken, asig xin
+    ibus = _bususe(itoken)
+    ga__buses[ibus] = asig
+endop
+
+opcode busmix, 0, ia
+    itoken, asig xin
+    ibus = _bususe(itoken)
+    ga__buses[ibus] = ga__buses[ibus] + asig
+endop
+
+schedule "__businit", 0, 1
+'''
+
+
+def _extractInstrNames(s:str) -> List[str]:
+    names = []
+    for line in s.splitlines():
+        if match := re.search(r"\binstr\s+\$\{(\w+)\}", line):
+            instrname = match.group(1)
+            names.append(instrname)
+    return names
+
+
+_instrNames = _extractInstrNames(_orc)
 
 # Constants
 CONSTS = {
@@ -446,16 +458,46 @@ CONSTS = {
     'reservedTablesStart': 500,
     'reservedInstrsStart': 500,
     'numReservedTables': 2000,
+    'maxNumInstrs': 10000
 }
 
-_tableNames = ['responses', 'subgains', 'busrefs']
+_tableNames = ['responses', 'subgains', 'tokenToInstrnum']
 
 BUILTIN_INSTRS = {k:i for i, k in enumerate(_instrNames, start=CONSTS['reservedInstrsStart'])}
-BUILTIN_INSTRS['cleanbuses'] = CONSTS['postProcInstrnum']
+BUILTIN_INSTRS['clearbuses'] = CONSTS['postProcInstrnum']
 BUILTIN_TABLES = {name:i for i, name in enumerate(_tableNames, start=1)}
 
 
-ORC_TEMPLATE = Template(_orc).safe_substitute(
-            **{name: f"{num} ; {name}" for name, num in BUILTIN_INSTRS.items()},
-            **BUILTIN_TABLES,
-            **CONSTS)
+@lru_cache(maxsize=0)
+def orcTemplate(busSupport=True) -> Template:
+    parts = [_orc]
+    if busSupport:
+        parts.append(_busOrc)
+    orc = "\n".join(parts)
+    subs: Dict[str, Any] = {name:f"{num} ; {name}"
+                            for name, num in BUILTIN_INSTRS.items()}
+    subs.update(BUILTIN_TABLES)
+    subs.update(CONSTS)
+    return Template(Template(orc).safe_substitute(**subs))
+
+
+
+def makeOrc(sr:int, ksmps:int, nchnls:int, nchnls_i:int,
+            backend:str, a4:float, globalcode:str="", includestr:str="",
+            numAudioBuses:int=32):
+    template = orcTemplate(busSupport=numAudioBuses>0)
+    return template.substitute(
+            sr=sr,
+            ksmps=ksmps,
+            nchnls=nchnls,
+            nchnls_i=nchnls_i,
+            backend=backend,
+            a4=a4,
+            globalcode=globalcode,
+            includes=includestr,
+            numAudioBuses=numAudioBuses)
+
+
+def busSupportCode(numAudioBuses:int, clearBusesInstrnum:int) -> str:
+    return Template(_busOrc).substitute(numAudioBuses=numAudioBuses,
+                                        clearbuses=f'{clearBusesInstrnum} ;  clearbuses')
