@@ -9,10 +9,12 @@ Example
 
     from csoundengine import *
     renderer = Renderer(sr=44100, nchnls=2)
+
     Instr('saw', r'''
       kmidi = p5
       outch 1, oscili:a(0.1, mtof:k(kfreq))
     ''').register(renderer)
+
     score = [('saw', 0,   2, 60),
              ('saw', 1.5, 4, 67),
              ('saw', 1.5, 4, 67.1)]
@@ -35,7 +37,9 @@ import textwrap as _textwrap
 from .config import config
 from . import csoundlib
 from .instr import Instr
-from . import tools
+from . import internalTools
+from . import engineorc
+
 from emlib import misc, iterlib
 import numpy as np
 
@@ -57,6 +61,36 @@ class ScoreEvent:
     dur: float
     args: List[float]
     eventId: int = 0
+    renderer: Opt[Renderer] = None
+
+    def setp(self, delay:float, *args, **kws) -> None:
+        """
+        Modify a parg of this synth.
+
+        Multiple pargs can be modified simultaneously. It only makes sense
+        to modify a parg if a k-rate variable was assigned to this parg
+        (see Renderer.setp for an example). A parg can be referred to via an integer,
+        corresponding to the p index (5 would refer to p5), or to the name
+        of the assigned k-rate variable as a string (for example, if there
+        is a line "kfreq = p6", both 6 and "kfreq" refer to the same parg).
+
+        Example
+        =======
+
+            >>> from csoundengine import *
+            >>> r = Renderer()
+            >>> Instr('sine', r'''
+            ... |kamp=0.1, kfreq=1000|
+            ... outch 1, oscili:ar(kamp, freq)
+            ... ''')
+            >>> event = r.sched('sine', 0, dur=4, pargs=[0.1, 440])
+            >>> event.setp(2, kfreq=880)
+            >>> event.setp(3, 'kfreq', 660, 'kamp', 0.5)
+
+        """
+        if not self.renderer:
+            raise RuntimeError("This ScoreEvent has no associated renderer")
+        self.renderer.setp(self, delay=delay, *args, **kws)
 
 
 class Renderer:
@@ -72,11 +106,37 @@ class Renderer:
         nchnls: number of channels
         ksmps: csound ksmps
         a4: reference frequency
-        maxpriorities: max. groups
+        maxpriorities: max. number of priority groups. This will determine
+            how long an effects chain can be
         bucketsize: max. number of instruments per priority group
 
+    Example
+    =======
+
+    .. code-block:: python
+
+        from csoundengine import *
+        renderer = Renderer(sr=44100, nchnls=2)
+
+        Instr('saw', r'''
+          kmidi = p5
+          outch 1, oscili:a(0.1, mtof:k(kfreq))
+        ''').register(renderer)
+
+        score = [('saw', 0,   2, 60),
+                 ('saw', 1.5, 4, 67),
+                 ('saw', 1.5, 4, 67.1)]
+        events = [renderer.sched(ev[0], delay=ev[1], dur=ev[2], pargs=ev[3:])
+                  for ev in score]
+
+        # offline events can be modified just like real-time events
+        renderer.automatep(events[0], 'kmidi', pairs=[0, 60, 2, 59])
+        renderer.setp(events[1], 3, 'kmidi', 67.2)
+        renderer.render("out.wav")
+
+
     """
-    _builtinInstrs = '''
+    _offlineOrc = '''
         instr _automatePargViaTable
           ip1 = p4
           ipindex = p5
@@ -113,12 +173,11 @@ class Renderer:
           endif
           turnoff
         endin
-
     '''
 
     def __init__(self, sr: int = None, nchnls: int = 2, ksmps: int = None,
-                 a4: float = None,
-                 maxpriorities=10, bucketsize=100):
+                 a4: float = None, maxpriorities=10, bucketsize=100,
+                 numAudioBuses=1000):
         """
 
         """
@@ -136,11 +195,14 @@ class Renderer:
         self._instrdefs: Dict[str, Instr] = {}
         self._instanceCounters: Dict[int, int] = {}
         self._numInstancesPerInstr = 10000
+        self._numAudioBuses = numAudioBuses
 
         # a list of i events, starting with p1
         self.events: List[List[float]] = []
         self.unscheduledEvents: List[List[float]] = []
-        self._csd.addGlobalCode(_textwrap.dedent(self._builtinInstrs))
+        self._csd.addGlobalCode(_textwrap.dedent(self._offlineOrc))
+        self._busSystemInitialized = False
+        self._busTokenCount = 0
 
     def _commitInstrument(self, instrname: str, priority=1) -> int:
         """
@@ -183,9 +245,79 @@ class Renderer:
     def registerInstr(self, instr: Instr) -> None:
         """
         Register an Instr to be used in this Renderer
+
+        Example
+        =======
+
+            >>> from csoundengine import *
+            >>> renderer = Renderer(sr=44100, nchnls=2)
+            >>> instrs = [
+            ... Instr('vco', r'''
+            ...   |kmidi=60|
+            ...   outch 1, vco2:a(0.1, mtof:k(kmidi))
+            ...   '''),
+            ... Instr('sine', r'''
+            ...   |kmidi=60|
+            ...   outch 1, oscili:a(0.1, mtof:k(kmidi))
+            ... ''')]
+            >>> for instr in instrs:
+            ...     renderer.registerInstr(instr)
+            >>> renderer.sched('vco', dur=4, kmidi=67)
+            >>> renderer.sched('sine', 2, dur=3, kmidi=68)
+            >>> renderer.render('out.wav')
+
         """
         self._instrdefs[instr.name] = instr
 
+    def defInstr(self, name: str, body: str, **kws) -> Instr:
+        """
+        Create an :class:`~csoundengine.instr.Instr` and register it with this renderer
+
+        Args:
+            name (str): the name of the created instr
+            body (str): the body of the instrument. It can have named
+                pfields (see example) or a table declaration
+            kws: any keywords are passed on to the Instr constructor.
+                See the documentation of Instr for more information.
+
+        Returns:
+            the created Instr. If needed, this instr can be registered
+            at any other Renderer/Session
+
+        Example
+        =======
+
+            >>> from csoundengine import *
+            >>> renderer = Renderer()
+            # An Instr with named pfields
+            >>> renderer.defInstr('synth', '''
+            ... |ibus, kamp=0.5, kmidi=60|
+            ... kfreq = mtof:k(lag:k(kmidi, 1))
+            ... a0 vco2 kamp, kfreq
+            ... a0 *= linsegr:a(0, 0.1, 1, 0.1, 0)
+            ... busout ibus, a0
+            ... ''')
+            # An instr with named table args
+            >>> renderer.defInstr('filter', '''
+            ... {bus=0, cutoff=1000, resonance=0.9}
+            ... a0 = busin(kbus)
+            ... a0 = moogladder2(a0, kcutoff, kresonance)
+            ... outch 1, a0
+            ... ''')
+
+            >>> bus = renderer.assignBus()
+            >>> synth = renderer.sched('sine', 0, dur=10, ibus=bus, kmidi=67)
+            >>> synth.setp(kmidi=60, delay=2)
+
+            >>> filt = session.sched('filter', 0, dur=synth.dur, priority=synth.priority+1,
+            ...                      tabargs={'bus': bus, 'cutoff': 1000})
+            >>> filt.automateTable('cutoff', [3, 1000, 6, 200, 10, 4000])
+            >>> bus.free()
+        """
+        instr = Instr(name=name, body=body, **kws)
+        self.registerInstr(instr)
+        return instr
+
     def registeredInstrs(self) -> KeysView:
         """
         Returns a seq. with the names of all registered Instrs
@@ -195,6 +327,13 @@ class Renderer:
     def addGlobalCode(self, code: str) -> None:
         """
         Add global code (instr 0)
+
+        Example
+        =======
+
+        >>> from csoundengine import *
+        >>> renderer = Renderer(...)
+        >>> renderer.addGlobalCode("giMelody[] fillarray 60, 62, 64, 65, 67, 69, 71")
         """
         self._csd.addGlobalCode(code)
 
@@ -212,7 +351,6 @@ class Renderer:
               priority=1,
               pargs: U[List[float], Dict[str, float]] = None,
               tabargs: Dict[str, float] = None,
-              unique: bool = True,
               **pkws) -> ScoreEvent:
         """
         Schedule an event
@@ -228,10 +366,31 @@ class Renderer:
             tabargs: a dict of the form param: value, to initialize
                 values in the exchange table (if defined by the given
                 instrument)
-            unique: if True, schedule a unique instance.
 
         Returns:
             a ScoreEvent, holding the csound event (p1, start, dur, args)
+
+
+        Example
+        =======
+
+            >>> from csoundengine import *
+            >>> renderer = Renderer(sr=44100, nchnls=2)
+            >>> instrs = [
+            ... Instr('vco', r'''
+            ...   |kmidi=60|
+            ...   outch 1, vco2:a(0.1, mtof:k(kmidi))
+            ...   '''),
+            ... Instr('sine', r'''
+            ...   |kamp=0.1, kmidi=60|
+            ...   outch 1, oscili:a(kamp, mtof:k(kmidi))
+            ... ''')]
+            >>> for instr in instrs:
+            ...     renderer.registerInstr(instr)
+            >>> renderer.sched('vco', dur=4, kmidi=67)
+            >>> renderer.sched('sine', 2, dur=3, kmidi=68)
+            >>> renderer.render('out.wav')
+
         """
         instr = self._instrdefs.get(instrname)
         if not instr:
@@ -242,14 +401,48 @@ class Renderer:
             tabnum = self._csd.addTableFromSeq(tableinit)
         else:
             tabnum = 0
-        args = tools.instrResolveArgs(instr, tabnum, pargs, pkws)
-        p1 = self._getUniqueP1(instrnum) if unique else instrnum
+        args = internalTools.instrResolveArgs(instr, tabnum, pargs, pkws)
+        p1 = self._getUniqueP1(instrnum)
         self._csd.addEvent(p1, start=delay, dur=dur, args=args)
         eventId = self._generateEventId()
         event = ScoreEvent(p1, delay, dur, args, eventId)
         self._eventsIndex[eventId] = event
         return event
 
+    def _initBusSystem(self) -> None:
+        if self._busSystemInitialized:
+            return
+        code = engineorc.busSupportCode(numAudioBuses=self._numAudioBuses,
+                                        clearBusesInstrnum=engineorc.CONSTS['postProcInstrnum'])
+        self._csd.addGlobalCode(code)
+        self._busSystemInitialized = True
+
+    def assignBus(self, kind='audio') -> int:
+        """
+        Assign a bus number
+
+        Example
+        =======
+
+            >>> from csoundengine.offline import Renderer
+            >>> r = Renderer()
+            >>> Instr('vco', r'''
+            ...     kfreq=p4
+            ...     kamp=p5
+            ...     asig vco2 1, kfreq
+            ...     aenv = linsegr:a(0, 0.01, 1, 0.01, 0)
+            ...     aenv *= lag:a(a(kamp), 0.1)
+            ...     asig *= aenv
+            ...     outch 1, asig
+            ... ''').register(r)
+        """
+        if kind != 'audio':
+            raise ValueError("Only audio buses are supported at the moment")
+        self._initBusSystem()
+        token = self._busTokenCount
+        self._busTokenCount += 1
+        return token
+
     def _generateEventId(self) -> int:
         out = self._idCounter
         self._idCounter += 1
@@ -265,7 +458,8 @@ class Renderer:
         Examples
         ========
 
-            >>> renderer = Renderer(...)
+            >>> from csoundengine.offline import Renderer
+            >>> renderer = Renderer()
             >>> instr = Instr("sine", ...)
             >>> renderer.registerInstr(instr)
             >>> renderer.sched("sine", ...)
@@ -302,15 +496,20 @@ class Renderer:
             outfile = tempfile.mktemp(suffix=".wav")
         if not self._csd.score:
             raise ValueError("score is empty")
-        kws = {}
         if quiet:
-            kws['supressdisplay'] = True
-            kws['piped'] = True
+            run_suppressdisplay = True
+            run_piped = True
+        else:
+            run_suppressdisplay = False
+            run_piped = False
+
         if samplefmt is None:
             ext = os.path.splitext(outfile)[1]
             samplefmt = csoundlib.bestSampleFormatForExtension(ext)
         self._csd.setSampleFormat(samplefmt)
-        proc = self._csd.run(output=outfile, **kws)
+        proc = self._csd.run(output=outfile,
+                             suppressdisplay=run_suppressdisplay,
+                             piped=run_piped)
         if openWhenDone:
             proc.wait()
             misc.open_with_standard_app(outfile, wait=True)
@@ -405,9 +604,48 @@ class Renderer:
                 idx = instr.pargIndex(k)
                 pairsd[idx] = v
         pairs = iterlib.flatdict(pairsd)
-        args = [event.p1, len(pairs)//2]
-        args.extend(pairs)
-        self._csd.addEvent("_pwrite", start=delay, dur=0.1, args=args)
+        pargs = [event.p1, len(pairs)//2]
+        pargs.extend(pairs)
+        self._csd.addEvent("_pwrite", start=delay, dur=0.1, args=pargs)
+
+    def readSoundfile(self, path: str, tabnum:int=None, chan=0, start=0., skiptime=0.) -> int:
+        """
+        Add code to this offline renderer to load a soundfile
+
+        Args:
+            path: the path of the soundfile to load
+            tabnum: the table number to assign, or None to autoassign a number
+            chan: the channel to read, or 0 to read all channels
+            start: moment in the score to read this soundfile
+            skiptime: skip this time at the beginning of the soundfile
+
+        Returns:
+            the assigned table number
+        """
+        return self._csd.addSndfile(sndfile=path, tabnum=tabnum,
+                                    start=start, skiptime=skiptime)
+
+    def playSample(self, tabnum:int, delay=0., chan=1, speed=1., gain=1., fade=0.,
+                   starttime=0., dur=-1) -> None:
+        """
+        Add an instrument definition and an event to play the given
+        table as sound (assumes that the table was allocated via
+        :meth:`~Renderer.readSoundFile` or any other GEN1 ftgen
+
+        Args:
+            tabnum: the table number to play
+            delay: when to start playback
+            chan: the channel to play
+            speed: the speed to play at
+            gain: apply a gain to playback
+            fade: fade-in / fade-out ramp, in seconds
+            starttime: playback does not start at the beginning of
+                the table but at `starttime`
+            dur: duration of playback. -1=until end of sample
+        """
+        self._csd.playTable(tabnum=tabnum, start=delay, dur=dur,
+                            gain=gain, speed=speed, chan=chan, fade=fade,
+                            skip=starttime)
 
     def automatep(self, event: ScoreEvent, param: str,
                   pairs: U[List[float], np.ndarray],
@@ -419,7 +657,7 @@ class Renderer:
             event: the event to automate, as returned by sched
             param (str): the name of the parameter to automate. The instr should
                 have a corresponding line of the sort "kparam = pn"
-            pairs: the automateion data as a flat list [t0, y0, t1, y1, ...], where
+            pairs: the automateion data as a flat list ``[t0, y0, t1, y1, ...]``, where
                 the times are relative to the start of the automation event
             mode (str): one of "linear", "cos", "smooth", "exp=xx" (see interp1d)
             delay: start time of the automation event. If None is given, the start
